import json
import pytest
import time

import numpy as np
from gameanalysis import rsgame

from egta import simsched


def test_basic_profile():
    with open('cdasim/game.json') as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    game = rsgame.emptygame_json(jgame)
    profs = game.random_profiles(20)
    cmd = ['python3', 'cdasim/sim.py', '--single', '1']

    with simsched.SimulationScheduler(game, conf, cmd) as sched:
        proms = [sched.schedule(p) for p in profs]
        pays = np.concatenate([p.get()[None] for p in proms])
    assert pays.shape == profs.shape
    assert np.allclose(pays[profs == 0], 0)
    assert np.any(pays != 0)


def test_delayed_fail():
    with open('cdasim/game.json') as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    game = rsgame.emptygame_json(jgame)
    prof = game.random_profiles()
    cmd = ['bash', '-c', 'sleep 1 && false']

    with simsched.SimulationScheduler(game, conf, cmd) as sched:
        with pytest.raises(RuntimeError):
            sched.schedule(prof).get()


def test_early_exit():
    with open('cdasim/game.json') as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    game = rsgame.emptygame_json(jgame)
    profs = game.random_profiles(20)
    cmd = ['bash', '-c', 'while read line; do : ; done']

    finished_scheduling = False
    with simsched.SimulationScheduler(game, conf, cmd) as sched:
        # Schedule promises but just exit
        for p in profs:
            sched.schedule(p)
        finished_scheduling = True
    assert finished_scheduling, \
        "didn't finish scheduling"


def test_read_delay_fail():
    with open('cdasim/game.json') as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    game = rsgame.emptygame_json(jgame)
    cmd = ['bash', '-c', 'read line && sleep 0.1 && false']

    scheduled = False
    with simsched.SimulationScheduler(game, conf, cmd) as sched:
        prom = sched.schedule(game.random_profiles())
        scheduled = True
        with pytest.raises(RuntimeError):
            prom.get()
    assert scheduled


def test_read_delay_schedule_fail():
    with open('cdasim/game.json') as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    game = rsgame.emptygame_json(jgame)
    cmd = ['bash', '-c', 'read line && sleep 0.2 && false']

    got_here = False
    with simsched.SimulationScheduler(game, conf, cmd) as sched:
        sched.schedule(game.random_profiles())
        # XXX For some reason the process hasn't always terminated after 3
        # seconds, but I can't afford to wait longer.
        time.sleep(3)  # make sure process is dead
        got_here = True
        with pytest.raises(RuntimeError):
            sched.schedule(game.random_profiles())
    assert got_here, \
        "didn't get to second schedule"


@pytest.mark.timeout(10)
def test_ignore_terminate_fail():
    with open('cdasim/game.json') as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    game = rsgame.emptygame_json(jgame)
    cmd = ['bash', '-c', 'trap "" SIGTERM && sleep 20']
    with simsched.SimulationScheduler(game, conf, cmd):
        # Wait for term to be captured
        time.sleep(1)


def test_json_decode_fail():
    with open('cdasim/game.json') as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    game = rsgame.emptygame_json(jgame)
    cmd = ['bash', '-c', 'echo "[" && while read line; do :; done']
    opened = False
    with simsched.SimulationScheduler(game, conf, cmd) as sched:
        time.sleep(1)
        opened = True
        with pytest.raises(RuntimeError):
            sched.schedule(game.random_profiles())
    assert opened



import pytest
import random
import time

import numpy as np
from egtaonline import mockapi
from gameanalysis import rsgame

from egta import countsched
from egta import eosched


def test_basic_profile():
    game = rsgame.emptygame([4, 4], [11, 11])
    profs = game.random_profiles(20)

    with mockapi.EgtaOnlineApi() as api:
        # Setup api
        sim = api.create_simulator(
            'sim', '1', delay_dist=lambda: random.random() / 10)
        sim.add_dict({role: strats for role, strats
                      in zip(game.role_names, game.strat_names)})

        # Schedule all new profiles and verify it works
        # This first time should have to wait to schedule more
        sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 1, 10, 0, 0)
        with sched:
            proms = [sched.schedule(p) for p in profs]
            pays = np.concatenate([p.get()[None] for p in proms])
        assert np.allclose(pays[profs == 0], 0)
        assert sched._num_running_profiles == 0

        # Schedule old profiles and verify it still works
        sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 1, 25, 0, 0)
        with sched:
            proms = [sched.schedule(p) for p in profs]
            pays = np.concatenate([p.get()[None] for p in proms])
        assert np.allclose(pays[profs == 0], 0)
        assert sched._num_running_profiles == 0

        # Schedule two at a time, in two batches
        base_sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 1, 25, 0, 0)
        sched = countsched.CountScheduler(base_sched, 2)
        with sched:
            proms = [sched.schedule(p) for p in profs]
            pays = np.concatenate([p.get()[None] for p in proms])
        assert np.allclose(pays[profs == 0], 0)
        assert base_sched._num_running_profiles == 0

        # Try again now that everything should be scheduled
        base_sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 1, 25, 0, 0)
        sched = countsched.CountScheduler(base_sched, 2)
        with sched:
            proms = [sched.schedule(p) for p in profs]
            pays = np.concatenate([p.get()[None] for p in proms])
        assert np.allclose(pays[profs == 0], 0)
        assert base_sched._num_running_profiles == 0


def test_extra_samples():
    game = rsgame.emptygame([4, 4], [11, 11])
    profs = game.random_profiles(20)

    with mockapi.EgtaOnlineApi() as api:
        # Setup api
        sim = api.create_simulator(
            'sim', '1', delay_dist=lambda: random.random() / 10)
        sim.add_dict({role: strats for role, strats
                      in zip(game.role_names, game.strat_names)})

        # Schedule all new profiles and verify it works
        # This first time should have to wait to schedule more
        sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 1, 25, 0, 0)
        with sched:
            proms = [sched.schedule(p) for p in profs]
            pays1 = np.concatenate([p.get()[None] for p in proms])
            proms = [sched.schedule(p) for p in profs]
            pays2 = np.concatenate([p.get()[None] for p in proms])
            pays3 = np.concatenate([p.get()[None] for p in proms])
        assert np.allclose(pays1[profs == 0], 0)
        assert np.allclose(pays2[profs == 0], 0)
        assert np.allclose(pays2, pays3)
        assert sched._num_running_profiles == 0


def test_existing_game():
    game = rsgame.emptygame([4, 4], [11, 11])
    profs = game.random_profiles(20)

    with mockapi.EgtaOnlineApi() as api:
        # Setup api
        sim = api.create_simulator(
            'sim', '1', delay_dist=lambda: random.random() / 10)
        sim.add_dict({role: strats for role, strats
                      in zip(game.role_names, game.strat_names)})
        eogame = sim.create_game('game', game.num_players)
        for role, count, strats in zip(game.role_names, game.num_role_players,
                                       game.strat_names):
            eogame.add_role(role, count)
            for strat in strats:
                eogame.add_strategy(role, strat)

        # Schedule all new profiles and verify it works
        # This first time should have to wait to schedule more
        sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 1, 25, 0, 0, game_id=eogame['id'])
        with sched:
            proms = [sched.schedule(p) for p in profs]
            pays1 = np.concatenate([p.get()[None] for p in proms])
            proms = [sched.schedule(p) for p in profs]
            pays2 = np.concatenate([p.get()[None] for p in proms])
        assert np.allclose(pays1[profs == 0], 0)
        assert np.allclose(pays2[profs == 0], 0)
        assert np.allclose(pays1, pays2)
        assert sched._num_running_profiles == 0


def test_exception_in_sechedule():
    game = rsgame.emptygame([4, 4], [11, 11])
    profs = game.random_profiles(20)

    with mockapi.ExceptionEgtaOnlineApi(TimeoutError, 60) as api:
        # Setup api
        sim = api.create_simulator(
            'sim', '1', delay_dist=lambda: random.random() / 10)
        sim.add_dict({role: strats for role, strats
                      in zip(game.role_names, game.strat_names)})

        sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 1, 25, 0, 0)
        opened = False
        with pytest.raises(TimeoutError):
            with sched:
                opened = True
                with pytest.raises(TimeoutError):
                    for p in profs:  # pragma: no branch
                        sched.schedule(p)
        assert opened


def test_exception_in_get():
    game = rsgame.emptygame([4, 4], [11, 11])
    profs = game.random_profiles(20)

    with mockapi.ExceptionEgtaOnlineApi(TimeoutError, 110) as api:
        # Setup api
        sim = api.create_simulator(
            'sim', '1', delay_dist=lambda: random.random() / 10)
        sim.add_dict({role: strats for role, strats
                      in zip(game.role_names, game.strat_names)})

        sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 1, 25, 0, 0)
        scheduled = False
        with pytest.raises(TimeoutError):
            with sched:
                proms = [sched.schedule(p) for p in profs]
                scheduled = True
                with pytest.raises(TimeoutError):
                    for p in proms:  # pragma: no branch
                        p.get()
        assert scheduled


def test_exception_pre_schedule():
    game = rsgame.emptygame([4, 4], [11, 11])
    prof = game.random_profiles()

    with mockapi.ExceptionEgtaOnlineApi(TimeoutError, 60) as api:
        # Setup api
        sim = api.create_simulator(  # pragma: no branch
            'sim', '1', delay_dist=lambda: random.random() / 10)
        sim.add_dict({role: strats for role, strats
                      in zip(game.role_names, game.strat_names)})

        sched = eosched.EgtaOnlineScheduler(
            api, sim['id'], game, 1, {}, 0.1, 25, 0, 0)
        slept = False
        with pytest.raises(TimeoutError):
            with sched:
                # so that enough calls to get_requirements are made
                time.sleep(1)
                slept = True
                with pytest.raises(TimeoutError):
                    sched.schedule(prof)
        assert slept



import json
import mock
import threading

import numpy as np
import pytest
from gameanalysis import gamegen
from gameanalysis import paygame
from gameanalysis import rsgame
from gameanalysis.reduction import deviation_preserving as dpr
from gameanalysis.reduction import twins as tr

from egta import innerloop
from egta import gamesched
from egta import simsched


SIMPLE_SIZES = [
    [[1], [1]],
    [[2], [2]],
    [[2, 2], [2, 2]],
    [[3, 2], [2, 3]],
    [[3, 4], [4, 3]],
]


def verify_dist_thresh(eqa, thresh=1e-3):
    for i, eqm in enumerate(eqa[:-1], 1):
        assert np.all(thresh ** 2 <= np.sum((eqm - eqa[i:]) ** 2, 1))


@pytest.mark.parametrize('players,strats', SIMPLE_SIZES)
def test_innerloop_simple(players, strats):
    sgame = paygame.samplegame_copy(gamegen.role_symmetric_game(
        players, strats))
    with gamesched.SampleGameScheduler(sgame) as sched:
        eqa = innerloop.inner_loop(sched, sgame)
    verify_dist_thresh(eqa)


@pytest.mark.parametrize('players,strats', SIMPLE_SIZES)
def test_innerloop_dpr(players, strats):
    redgame = rsgame.emptygame(players, strats)
    fullgame = rsgame.emptygame(redgame.num_role_players ** 2,
                                redgame.num_role_strats)
    profs = dpr.expand_profiles(fullgame, redgame.all_profiles())
    pays = np.random.random(profs.shape)
    pays[profs == 0] = 0
    sgame = paygame.samplegame_replace(fullgame, profs, [pays[:, None]])
    with gamesched.SampleGameScheduler(sgame) as sched:
        eqa = innerloop.inner_loop(sched, sgame, dpr, redgame.num_role_players)
    verify_dist_thresh(eqa)


@pytest.mark.parametrize('players,strats', SIMPLE_SIZES)
def test_innerloop_by_role_simple(players, strats):
    sgame = paygame.samplegame_copy(gamegen.role_symmetric_game(
        players, strats))
    with gamesched.SampleGameScheduler(sgame) as sched:
        eqa = innerloop.inner_loop(sched, sgame, devs_by_role=True)
    verify_dist_thresh(eqa)


@pytest.mark.parametrize('when', ['schedule', 'pre-get', 'post-get'])
@pytest.mark.parametrize('count', [1, 5, 10])
@pytest.mark.parametrize('players,strats',
                         [[[2, 3], [3, 2]], [[4, 3], [3, 4]]])
def test_innerloop_failures(players, strats, count, when):
    game = gamegen.role_symmetric_game(players, strats)
    with ExceptionScheduler(game, count, when) as sched:
        with pytest.raises(SchedulerException):
            innerloop.inner_loop(sched, game, subgame_size=5)


@pytest.mark.parametrize('count', [1, 5, 10])
@pytest.mark.parametrize('players,strats',
                         [[[2, 3], [3, 2]], [[4, 3], [3, 4]]])
def test_threading_failures(players, strats, count):
    game = gamegen.role_symmetric_game(players, strats)
    lock = threading.Lock()
    num = [count]
    old_thread = threading.Thread

    def fail_in(*args, **kwargs):
        with lock:
            num[0] -= 1
            if num[0] <= 0:
                return ExceptionThread(*args, **kwargs)
            else:
                return old_thread(*args, **kwargs)

    with gamesched.RsGameScheduler(game) as sched:
        with mock.patch('threading.Thread', side_effect=fail_in):
            with pytest.raises(ThreadException):
                innerloop.inner_loop(sched, game, subgame_size=5)


@pytest.mark.parametrize('eq_prob', [x / 10 for x in range(11)])
def test_innerloop_known_eq(eq_prob):
    sgame = paygame.samplegame_copy(gamegen.sym_2p2s_known_eq(eq_prob))
    with gamesched.SampleGameScheduler(sgame) as sched:
        eqa = innerloop.inner_loop(sched, sgame, devs_by_role=True)
    assert eqa.size, "didn't find equilibrium"
    expected = [eq_prob, 1 - eq_prob]
    assert np.isclose(eqa, expected, atol=1e-3, rtol=1e-3).all(-1).any()
    verify_dist_thresh(eqa)


@pytest.mark.parametrize('players,strats', SIMPLE_SIZES)
@pytest.mark.parametrize('num', [1, 2])
def test_innerloop_num_eqa(players, strats, num):
    sgame = paygame.samplegame_copy(
        gamegen.role_symmetric_game(players, strats))
    with gamesched.SampleGameScheduler(sgame) as sched:
        eqa = innerloop.inner_loop(
            sched, sgame, num_equilibria=num, devs_by_role=True)
    verify_dist_thresh(eqa)


def test_backups_used():
    """Test that outerloop uses backups

    Since subgame size is 1, but the only equilibria has support two, this must
    use backups to find an equilibrium."""
    sgame = paygame.samplegame_copy(gamegen.sym_2p2s_known_eq(0.5))
    with gamesched.SampleGameScheduler(sgame) as sched:
        eqa = innerloop.inner_loop(sched, sgame, subgame_size=1)
    assert eqa.size, "didn't find equilibrium"
    expected = [0.5, 0.5]
    assert np.isclose(eqa, expected, atol=1e-3, rtol=1e-3).all(-1).any()
    verify_dist_thresh(eqa)


@pytest.mark.long
def test_simsched():
    with open('cdasim/game.json') as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    game = rsgame.emptygame_json(jgame)
    cmd = ['python3', 'cdasim/sim.py', '--single', '1']
    with simsched.SimulationScheduler(game, conf, cmd) as sched:
        innerloop.inner_loop(sched, game, tr)


class SchedulerException(Exception):
    """Exception to be thrown by ExceptionScheduler"""
    pass


class ExceptionScheduler(gamesched.RsGameScheduler):
    """Scheduler that allows triggering exeptions on command"""

    def __init__(self, game, error_after, call_type):
        super().__init__(game)
        self._calls = 0
        self._error_after = error_after
        self._call_type = call_type
        self._lock = threading.Lock()

    def _inc_and_check(self, name):
        if self._call_type == name:
            with self._lock:
                self._calls += 1
                if self._error_after <= self._calls:
                    raise SchedulerException

    def schedule(self, profile):
        self._inc_and_check('schedule')
        return ExceptionPromise(self, super().schedule(profile))


class ExceptionPromise(object):
    """Promise for ExceptionScheduler"""

    def __init__(self, sched, prom):
        self._sched = sched
        self._prom = prom

    def get(self):
        self._sched._inc_and_check('pre-get')
        res = self._prom.get()
        self._sched._inc_and_check('post-get')
        return res


class ThreadException(Exception):
    """Exception to be thrown by ExceptionScheduler"""
    pass


class ExceptionThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def start(self, *args, **kwargs):
        raise ThreadException()



import argparse
import json
import logging

import numpy as np
from gameanalysis import nash
from gameanalysis import paygame
from gameanalysis import reduction
from gameanalysis import regret
from gameanalysis import rsgame
from gameanalysis import subgame


_log = logging.getLogger(__name__)


def add_parser(subparsers):
    parser = subparsers.add_parser(
        'brute', help="""Compute equilibria by sampling all profiles""",
        description="""Samples profiles from the entire game, and then runs
        standard equilibrium finding. For games with a large number of players,
        a reduction should be specified. A list of is returned where each
        element has an "equilibrium" and the corresponding "regret" in the full
        game.""")
    parser.add_argument(
        '--regret-thresh', metavar='<reg>', type=float, default=1e-3,
        help="""Regret threshold for a mixture to be considered an equilibrium.
        (default: %(default)g)""")
    parser.add_argument(
        '--dist-thresh', metavar='<norm>', type=float, default=1e-3,
        help="""Norm threshold for two mixtures to be considered distinct.
        (default: %(default)g)""")
    parser.add_argument(
        '--subgame', '-s', metavar='<subgame-file>',
        type=argparse.FileType('r'), help="""Specify an optional subgame to
        sample instead of the whole game. Only deviations from the subgame will
        be scheduled.""")
    reductions = parser.add_mutually_exclusive_group()
    reductions.add_argument(
        '--dpr', metavar='<role:count,role:count,...>', help="""Specify a
        deviation preserving reduction.""")
    reductions.add_argument(
        '--hr', metavar='<role:count,role:count,...>', help="""Specify a
        hierarchical reduction.""")
    return parser


def parse_reduction(game, red):
    reduced_players = np.empty(game.num_roles, int)
    for role_red in red.strip().split(','):
        role, count = role_red.strip().split(':')
        reduced_players[game.role_index(role.strip())] = int(count)
    return reduced_players


def run(scheduler, game, args):
    game = rsgame.emptygame_copy(game)

    if args.dpr is not None:
        red_players = parse_reduction(game, args.dpr)
        red = reduction.deviation_preserving
    elif args.hr is not None:
        red_players = parse_reduction(game, args.hr)
        red = reduction.hierarchical
    else:
        red = reduction.identity
        red_players = None

    if args.subgame is None:
        sub = np.ones(game.num_strats, bool)
    else:
        sub = game.from_subgame_json(json.load(args.subgame))

    subg = game.subgame(sub)
    devprofs = red.expand_deviation_profiles(game, sub, red_players)
    subprofs = subgame.translate(
        red.expand_profiles(subg, red.reduce_game(
            subg, red_players).all_profiles()), sub)
    profiles = np.concatenate([subprofs, devprofs])
    promises = [scheduler.schedule(prof) for prof in profiles]
    payoffs = np.concatenate([prom.get()[None] for prom in promises])
    game = red.reduce_game(paygame.game_replace(
        game, profiles, payoffs), red_players)
    eqa = subgame.translate(nash.mixed_nash(
        game.subgame(sub), regret_thresh=args.regret_thresh,
        dist_thresh=args.dist_thresh), sub)
    reg_info = []
    for eqm in eqa:
        gains = regret.mixture_deviation_gains(game, eqm)
        bri = np.argmax(gains)
        reg_info.append((gains[bri],) + game.role_strat_names[bri])

    _log.error("brute sampling finished finding %d equilibria:\n%s",
               eqa.shape[0], '\n'.join(
                   '{:d}) {} with regret {:g} to {} {}'.format(
                       i, game.to_mix_repr(eqm), reg, role, strat)
                   for i, (eqm, (reg, role, strat))
                   in enumerate(zip(eqa, reg_info), 1)))

    json.dump([{'equilibrium': game.to_mix_json(eqm),
                'regret': reg,
                'best_response': {'role': role, 'strat': strat}}
               for eqm, (reg, role, strat)
               in zip(eqa, reg_info)], args.output)
    args.output.write('\n')



import pytest

from gameanalysis import gamegen
from gameanalysis import paygame

from egta import gamesched
from egta import savesched


@pytest.mark.parametrize('_', range(20))
def test_basic_profile(_):
    sgame = paygame.samplegame_copy(
        gamegen.role_symmetric_game([4, 3], [3, 4]))
    profs = sgame.all_profiles()

    sched = savesched.SaveScheduler(
        sgame, gamesched.SampleGameScheduler(sgame))
    with sched:
        proms = [sched.schedule(p) for p in profs]
        for p in proms:
            p.get()
        for p in proms:
            p.get()
    savegame = sched.get_samplegame()
    assert sgame == savegame



import numpy as np
from gameanalysis import gamegen

from egta import gamesched
from egta import countsched


def test_basic_profile():
    sgame = gamegen.add_noise(
        gamegen.role_symmetric_game([4, 3], [3, 4]), 1, 3)
    profs = sgame.random_profiles(20)

    sched = countsched.CountScheduler(gamesched.SampleGameScheduler(sgame), 10)
    with sched:
        proms = [sched.schedule(p) for p in profs]
        pays = np.concatenate([p.get()[None] for p in proms])
        pays2 = np.concatenate([p.get()[None] for p in proms])
    assert np.allclose(pays[profs == 0], 0)
    assert np.allclose(pays, pays2)



import itertools
import json
import pytest
import subprocess
import tempfile
from os import path

from gameanalysis import rsgame


DIR = path.dirname(path.realpath(__file__))
EGTA = path.join(DIR, '..', 'bin', 'egta')
SIM_DIR = path.join(DIR, '..', 'cdasim')
SMALL_GAME = path.join(SIM_DIR, 'small_game.json')
GAME = path.join(SIM_DIR, 'game.json')
DATA_GAME = path.join(SIM_DIR, 'data_game.json')
SIM = [path.join(DIR, '..', 'bin', 'python'), path.join(SIM_DIR, 'sim.py'),
       '1', '--single']


def run(inp, *args):
    res = subprocess.run((EGTA,) + args, input=inp.encode('utf-8'),
                         stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return (not res.returncode, res.stdout.decode('utf-8'),
            res.stderr.decode('utf-8'))


def assert_term(sleep1, sleep2, *args):
    proc = subprocess.Popen((EGTA,) + args)
    try:
        proc.wait(sleep1)
        proc.kill()
        assert False, "process finished during first sleep"
    except subprocess.TimeoutExpired:
        pass
    proc.terminate()
    try:
        assert proc.wait(sleep2), "process succeeded after termination"
    except subprocess.TimeoutExpired:
        proc.kill()
        assert False, "process didn't terminate in second sleep"


def test_help():
    succ, _, _ = run('')
    assert not succ
    succ, _, _ = run('', '--fail')
    assert not succ
    succ, _, err = run('', '--help')
    assert succ, err
    succ, _, err = run('', 'brute', '--help')
    assert succ, err
    succ, _, err = run('', 'brute', 'game', '--help')
    assert succ, err
    succ, _, err = run('', 'brute', 'sim', '--help')
    assert succ, err
    succ, _, err = run('', 'brute', 'eo', '--help')
    assert succ, err
    succ, _, err = run('', 'quiesce', '--help')
    assert succ, err


def test_brute_game():
    succ, out, err = run(
        '', '--count', '2', '--game-json', DATA_GAME, 'brute', 'game')
    assert succ, err
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_brute_game_tag():
    succ, out, err = run(
        '', '--tag', 'test', '--count', '2', '--game-json', DATA_GAME, 'brute',
        'game')
    assert succ, err
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_brute_game_subgame():
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    sub = json.dumps(reader.to_subgame_json(reader.random_subgames()))
    succ, out, err = run(
        sub, '--count', '2', '--game-json', DATA_GAME, 'brute', '--subgame',
        '-', 'game')
    assert succ, err
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_brute_game_term():
    assert_term(0.5, 0.5, '--count', '2', '--game-json', DATA_GAME, 'brute',
                'game')


def test_brute_dpr_game():
    succ, out, err = run(
        '', '--count', '2', '--game-json', DATA_GAME, 'brute', '--dpr',
        'buyers:2,sellers:2', 'game')
    assert succ, err
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_brute_hr_game():
    succ, out, err = run(
        '', '--count', '2', '--game-json', DATA_GAME, 'brute', '--hr',
        'buyers:2,sellers:2', 'game')
    assert succ, err
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_prof_data():
    succ, out, err = run(
        '', '--profile-data', '/dev/null', '--game-json', DATA_GAME, 'brute',
        'game', '--sample')
    assert succ, err
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_sim():
    succ, out, err = run(
        '', '--game-json', SMALL_GAME, 'brute', 'sim', '--', *SIM)
    assert succ, err
    with open(SMALL_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_brute_sim_term():
    assert_term(0.5, 0.5, '--game-json', SMALL_GAME, 'brute', 'sim', '--',
                *SIM)


def test_sim_delayed_fail():
    succ, _, _ = run(
        '', '--game-json', SMALL_GAME, 'brute', 'sim', '--', 'bash', '-c',
        'sleep 1 && false')
    assert not succ


def test_sim_read_delayed_fail():
    succ, _, _ = run(
        '', '--game-json', SMALL_GAME, 'brute', 'sim', '--', 'bash', '-c',
        'read line && false')
    assert not succ


def test_sim_conf():
    with open(SMALL_GAME) as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    conf['markup'] = 'standard'
    reader = rsgame.emptygame_json(jgame)
    with tempfile.NamedTemporaryFile('w') as conf_file:
        json.dump(conf, conf_file)
        conf_file.flush()
        succ, out, err = run(
            '', '--game-json', SMALL_GAME, 'brute', 'sim', '-c',
            conf_file.name, '--', *SIM)
    assert succ, err
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_innerloop():
    succ, out, err = run('', '--game-json', DATA_GAME, 'quiesce', 'game')
    assert succ, err
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_innerloop_game_term():
    assert_term(0.5, 0.5, '--game-json', DATA_GAME, 'quiesce', 'game')


def test_innerloop_sim_term():
    assert_term(0.5, 0.5, '--game-json', SMALL_GAME, 'quiesce', 'sim', '--',
                *SIM)


def test_innerloop_dpr():
    succ, out, err = run(
        '', '--game-json', DATA_GAME, 'quiesce', '--dpr', 'buyers:2,sellers:2',
        'game')
    assert succ, err
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


def test_innerloop_hr():
    succ, out, err = run(
        '', '--game-json', DATA_GAME, 'quiesce', '--hr', 'buyers:2,sellers:2',
        'game')
    assert succ, err
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


@pytest.mark.egta
def test_game_id_brute_egta_game():
    succ, out, err = run(
        '', '-g1466', 'brute', '--dpr', 'buyers:2,sellers:2', 'eo', '2048',
        '60')
    assert succ, err
    with open(SMALL_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    for eqm in json.loads(out):
        reader.from_mix_json(eqm['equilibrium'])


@pytest.mark.egta
def test_game_id_brute_egta_game_wconf():
    with open(SMALL_GAME) as f:
        jgame = json.load(f)
    conf = jgame['configuration']
    reader = rsgame.emptygame_json(jgame)
    with tempfile.NamedTemporaryFile('w') as conf_file:
        json.dump(conf, conf_file)
        conf_file.flush()
        succ, out, err = run(
            '', '-g1466', 'brute', '--dpr', 'buyers:2,sellers:2', 'eo', '-c',
            conf_file.name, '2048', '60')
    assert succ, err
    for eqm in out[:-1].split('\n'):
        reader.from_mix_json(json.loads(eqm))


def test_boot_game():
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    mix = reader.random_mixtures()
    with tempfile.NamedTemporaryFile('w') as mix_file:
        json.dump(reader.to_mix_json(mix), mix_file)
        mix_file.flush()
        succ, out, err = run(
            '', '--game-json', DATA_GAME, 'boot', mix_file.name, '10',
            'game')
    assert succ, err
    results = json.loads(out)
    assert {'total', 'buyers', 'sellers'} == results.keys()
    for val in results.values():
        assert {'surplus', 'regret', 'response'} == val.keys()


def test_boot_game_percs():
    with open(DATA_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    mix = reader.random_mixtures()
    succ, out, err = run(
        json.dumps(reader.to_mix_json(mix)), '--game-json', DATA_GAME,
        'boot', '-', '20', '--percentiles', '95,99', 'game')
    assert succ, err
    results = json.loads(out)
    assert {'mean', '99', '95'} == results.keys()
    for val in results.values():
        assert {'total', 'buyers', 'sellers'} == val.keys()
    for val in results['mean'].values():
        assert {'surplus', 'regret', 'response'} == val.keys()
    for val in itertools.chain(results['95'].values(), results['99'].values()):
        assert {'surplus', 'regret'} == val.keys()


def test_boot_sim():
    with open(SMALL_GAME) as f:
        reader = rsgame.emptygame_json(json.load(f))
    mix = reader.random_mixtures()
    succ, out, err = run(
        json.dumps(reader.to_mix_json(mix)), '--game-json', SMALL_GAME,
        'boot', '-', '50', '--chunk-size', '10', 'sim', '--', *SIM)
    assert succ, err
    results = json.loads(out)
    assert {'total', 'buyers', 'sellers'} == results.keys()
    for val in results.values():
        assert {'surplus', 'regret', 'response'} == val.keys()



import numpy as np
import numpy.random as rand
from gameanalysis import gamegen
from gameanalysis import agggen

from egta import gamesched


def test_basic_profile():
    game = gamegen.role_symmetric_game([4, 3], [3, 4])
    profs = game.random_profiles(20)

    with gamesched.RsGameScheduler(game) as sched:
        proms = [sched.schedule(p) for p in profs]
        pays = np.concatenate([p.get()[None] for p in proms])
    assert np.allclose(pays[profs == 0], 0)


def test_basic_profile_sample():
    sgame = gamegen.add_noise(
        gamegen.role_symmetric_game([4, 3], [3, 4]), 1, 3)
    profs = sgame.random_profiles(20)

    with gamesched.SampleGameScheduler(sgame) as sched:
        proms = [sched.schedule(p) for p in profs]
        pays = np.concatenate([p.get()[None] for p in proms])
    assert np.allclose(pays[profs == 0], 0)


def test_duplicate_profile_sample():
    sgame = gamegen.add_noise(
        gamegen.role_symmetric_game([4, 3], [3, 4]), 1, 1)
    profs = sgame.random_profiles(20)

    with gamesched.SampleGameScheduler(sgame) as sched:
        proms = [sched.schedule(p) for p in profs]
        pays1 = np.concatenate([p.get()[None] for p in proms])
        proms = [sched.schedule(p) for p in profs]
        pays2 = np.concatenate([p.get()[None] for p in proms])
    assert np.allclose(pays1[profs == 0], 0)
    assert np.allclose(pays2[profs == 0], 0)
    assert np.allclose(pays1, pays2)


def test_basic_profile_aggfn():
    agame = agggen.random_aggfn([4, 3], [3, 4], 5)
    profs = agame.random_profiles(20)

    with gamesched.RsGameScheduler(agame) as sched:
        proms = [sched.schedule(p) for p in profs]
        pays = np.concatenate([p.get()[None] for p in proms])
    assert np.allclose(pays[profs == 0], 0)


def test_noise_profile():
    sgame = gamegen.add_noise(
        gamegen.role_symmetric_game([4, 3], [3, 4]), 1, 3)
    profs = sgame.random_profiles(20)

    sched = gamesched.SampleGameScheduler(
        sgame, lambda w: rand.normal(0, w, sgame.num_strats),
        lambda: (rand.random(),))
    with sched:
        proms = [sched.schedule(p) for p in profs]
        pays = np.concatenate([p.get()[None] for p in proms])
    assert np.allclose(pays[profs == 0], 0)


def test_duplicate_prof():
    game = gamegen.role_symmetric_game([4, 3], [3, 4])
    profs = game.random_profiles(20)

    with gamesched.RsGameScheduler(game) as sched:
        proms = [sched.schedule(p) for p in profs]
        pays = np.concatenate([p.get()[None] for p in proms])
        assert np.allclose(pays[profs == 0], 0)

        proms = [sched.schedule(p) for p in profs]
        pays = np.concatenate([p.get()[None] for p in proms])
        assert np.allclose(pays[profs == 0], 0)
